/* Y o u r   D e s c r i p t i o n                       */
/*                            AppBuilder Photon Code Lib */
/*                                         Version 1.11  */

/* Standard headers */
#include <time.h>
#include <stdio.h>
#include <dirent.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>

/* Toolkit headers */
#include <Ph.h>
#include <Pt.h>
#include <Ap.h>

/* Local headers */
#include "abimport.h"
#include "proto.h"

/*
 *	The IS_NEW_DIR flag will be used to indicate
 *	that an item is a directory but we haven't read it yet
 */
#define IS_NEW_DIR 0x800

#if IS_NEW_DIR & Pt_TREE_ITEM_USED_FLAGS
	#error The IS_NEW_DIR flag conflicts with something
#endif

/*
 *	The "user data" attached to a tree item:
 */
struct TreeItem {
	struct stat fstat;
	char fullpath[1];  /* We will allocate more than that */
	};

/*
 *	Indexes and names for tree images
 */
enum {
	ICON_OPENDIR = 0, ICON_CLOSEDDIR, ICON_EXEC,
	ICON_OTHER, ICON_ERROR
	};

const char *const icon_names[] = {
	"icon_opendir", "icon_closeddir", "icon_exec",
	"icon_other", "icon_error"
	};

#define NUMBER_OF_ICONS ( sizeof(icon_names) / sizeof(icon_names[0]) )

void nomemory( void ) {
	fputs( "No memory\n", stderr );
	exit( EXIT_FAILURE );
	}

void *allocmem( size_t nbytes ) {
	void *result;
	if ( ( result = malloc( nbytes ) ) == NULL )
		nomemory();
	return result;
	}

/*
 *	Check the file and allocate a tree item
 */
PtTreeItem_t *makeitem( const char *name, PtTreeItem_t *parent ) {
	int len; PtTreeItem_t *item = NULL;
	struct TreeItem *data, *pdata;
	char string[ PATH_MAX + 60 ], date[ 40 ];
	int sel, unsel, flag = 0;
	if ( parent == NULL )
		len = 0;
	else {
		pdata = parent->data;
		if	(	( len = strlen( pdata->fullpath ) ) != 0
			&&	pdata->fullpath[ len-1 ] == '/'
				)
			--len;
		}
	data = allocmem( len + sizeof(*data) + 1 + strlen(name) );
	if ( parent != NULL ) {
		strcpy( data->fullpath, pdata->fullpath );
		data->fullpath[ len++ ] = '/';
		}
	strcpy( data->fullpath + len, name );
	if ( stat( data->fullpath, &data->fstat ) == -1 ) {
		int err = errno;
		memset( &data->fstat, 0, sizeof(data->fstat) );
		sprintf( string, "%s\t%s\tError %d", name, strerror(err), err );
		sel = unsel = ICON_ERROR;
		}
	else {
		strftime( date, sizeof(date), "%c",
			localtime( & data->fstat.st_mtime )
			);
		sprintf( string, "%s\t%s\t%lu",
			name, date, (unsigned long) data->fstat.st_size
			);

		if ( S_ISDIR( data->fstat.st_mode ) ) {
			flag = Pt_TREE_ITEM_EXPANDABLE | IS_NEW_DIR;
			sel   = ICON_OPENDIR;
			unsel = ICON_CLOSEDDIR;
			}
		else
			if ( data->fstat.st_mode & ( S_IXUSR | S_IXGRP | S_IXOTH ) )
				sel = unsel = ICON_EXEC;
			else
				sel = unsel = ICON_OTHER;
		}
	if ( ( item = PtTreeAllocItem( ABW_tree, string, sel, unsel ) ) == NULL )
		nomemory();
	item->gen.list.flags |= flag;
	item->data = data;
	return item;
	}

int
base_setup( PtWidget_t *link_instance, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
	{
	PtTreeItem_t *item;
	ApDBase_t *db;
	PtArg_t arg;

	/* Setup tree images */
	if ( ( db = ApOpenDBase( ABM_tree_icons ) ) != NULL ) {
		int i; PhImage_t *img;
		for ( i = 0; i != NUMBER_OF_ICONS; ++i )
			if ( ( img = ApGetImageRes( db, icon_names[i] ) ) != NULL )
				PtTreeAddImages( ABW_tree, img, 1 );
			else
				break;
		}

	/*	Set the Pt_ARG_TREE_IMGMASK resource
	 *	(it isn't supported by PhAB in Photon 1.1)
	 */
	PtSetArg( &arg, Pt_ARG_TREE_IMGMASK, Pt_TREE_ITEM_EXPANDED, 0 );
	PtSetResources( ABW_tree, 1, &arg );

	/* Create the first item */
	if ( ( item = makeitem( "/", NULL ) ) != NULL )
		PtTreeAddFirst( ABW_tree, item, NULL );

	/* eliminate 'unreferenced' warnings */
	link_instance = link_instance, apinfo = apinfo, cbinfo = cbinfo;
	return( Pt_CONTINUE );
	}

/*
 *	Pt_CB_TREE_STATE callback function
 */
int
item_state( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
	{
	PtTreeCallback_t *cb = cbinfo->cbdata;
	PtTreeItem_t *parent = cb->item;
	struct TreeItem *data = parent->data;
	DIR *dp;
	if	(	cbinfo->reason_subtype == Pt_TREE_EXPANDING
		&&	parent->gen.list.flags & IS_NEW_DIR
			)
		if ( ( dp = opendir( data->fullpath ) ) == NULL )
			cb->expand = Pt_END;
		else {
			struct dirent *de;
			if ( ( de = readdir( dp ) ) != NULL ) {
				PtTreeItem_t *child, *prev;
				prev = makeitem( de->d_name, parent );
				PtTreeAddFirst( ABW_tree, prev, parent );
				while ( ( de = readdir( dp ) ) != NULL ) {
					child = makeitem( de->d_name, parent );
					PtTreeAddAfter( ABW_tree, child, prev );
					prev = child;
				}	}
			closedir( dp );
			parent->gen.list.flags &= ~IS_NEW_DIR;
			}

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;
	return( Pt_CONTINUE );
	}

